from __future__ import annotations
from typing import Dict, Iterable, Optional

import shlex
import subprocess

__all__ = ["Engine"]


class Engine:
    def __init__(
        self,
        executable: str,
        num_processors: Optional[int] = None,
        num_threads: Optional[int] = None,
        arguments: Optional[Iterable[str]] = None,
    ) -> None:

        self.executable = executable
        self.num_processors = num_processors
        self.num_threads = num_threads
        self.arguments = arguments or []

    def env(self) -> Dict[str, str]:
        return None

    def command(
        self,
        inp: str,
        out: str,
        work_dir: str,
        arguments: Optional[Iterable[str]] = None,
    ) -> str:
        args = list(self.arguments) + list(arguments or [])
        arg_str = " ".join(args)
        # FIXME: shlex.quote should be used but it doesn't work...
        return shlex.split(f"{self.executable} {arg_str}")

    def execute(self, io: materia.IO, arguments: Optional[Iterable[str]] = None) -> str:
        with io() as _io:
            cmd = self.command(_io.inp, _io.out, _io.work_dir, arguments)
            with open(_io.inp, "r") as inp:
                with open(_io.out, "w") as out:
                    env = self.env()
                    if env is None:
                        subprocess.call(cmd, stdin=inp, stdout=out, cwd=io.work_dir)
                    else:
                        subprocess.call(
                            cmd, stdin=inp, stdout=out, env=self.env(), cwd=io.work_dir
                        )

            with open(_io.out, "r") as f:
                return "".join(f.readlines())



from __future__ import annotations
from typing import Dict, Optional

import copy

from .action import Action

__all__ = ["InsertTasks", "Modify", "Rerun"]


class InsertTasks(Action):
    def __init__(self, *tasks, requires_kw=None) -> None:
        # NOTE: the first task in tasks must be the head, and the last task in tasks must be the tail!
        self.tasks = copy.deepcopy(tasks)
        self.requires_kw = requires_kw

    def run(self, node, tasks, links, done: Dict[int, bool]) -> None:
        head_task_node = len(tasks)
        # this value is safe to use in the scope of this method because only the producer runs actions, and there is only one producer at a time
        tail_task_node = head_task_node + len(self.tasks) - 1
        # this value is safe to use in the scope of this method because only the producer runs actions, and there is only one producer at a time
        tasks.extend(self.tasks)
        for k, v in links.items():
            links[k] = [(kw, x) if x != node else (kw, tail_task_node) for kw, x in v]
        links[head_task_node] = [(self.requires_kw, node)]
        for i in range(head_task_node, tail_task_node + 1):
            done[i] = False


class Modify(Action):
    def modify(self, task: materia.Task) -> materia.Task:
        # NOTE: this should return the modified task - this can be a copy or it can modify task in place and return it, either way
        raise NotImplementedError

    def run(self, node, tasks, links, done: Dict[int, bool]) -> None:
        tasks[node] = self.modify(task=tasks[node])


class Rerun(Action):
    def run(self, node, tasks, links, done: Dict[int, bool]) -> None:
        done[node] = False



from __future__ import annotations
from typing import Iterable, Optional

import shlex

from .engine import Engine

__all__ = ["Openbabel"]


class Openbabel(Engine):
    def __init__(
        self,
        executable: Optional[str] = "obabel",
        arguments: Optional[Iterable[str]] = None,
    ) -> None:
        super().__init__(executable=executable, arguments=arguments)

    def command(
        self,
        inp: str,
        out: str,
        work_dir: str,
        arguments: Optional[Iterable[str]] = None,
    ) -> str:
        args = list(self.arguments) + list(arguments or [])
        arg_str = " ".join(args)
        # FIXME: shlex.quote should be used but it doesn't work...
        return shlex.split(f"{self.executable} {inp} {arg_str}")


# from __future__ import annotations
# import contextlib
# import os
# import mtr
# import re
# import tempfile
# from typing import Iterable, Optional, Union

# from ...workflow.tasks.task import Task

# __all__ = [
#     "OpenbabelConvertToFile",
#     "OpenbabelConvertToMol",
#     "OpenbabelConvertToPDB",
#     "OpenbabelConvertToSMILES",
#     "OpenbabelConvertToSDF",
# ]


# class OpenbabelConvertToFile(Task):
#     def __init__(
#         self,
#         engine: mtr.OpenbabelEngine,
#         filetype: str,
#         output_name: Optional[str] = None,
#         log_name: Optional[str] = "obabel.log",
#         work_dir: Optional[str] = ".",
#         keep_logs: bool = True,
#         handlers: Optional[Iterable[mtr.Handler]] = None,
#         name: Optional[str] = None,
#     ) -> None:
#         super().__init__(handlers=handlers, name=name)
#         self.engine = engine
#         self.filetype = filetype
#         self.output_name = output_name
#         self.log_name = log_name
#         self.work_dir = work_dir
#         self.keep_logs = keep_logs

#     def run(self, structure: Union[str, mtr.Structure]) -> str:
#         with mtr.work_dir(self.work_dir) as wd:
#             arguments = [f"-o{self.filetype}"]

#             if self.output_name is not None:
#                 output_path = (
#                     f"{mtr.expand(self.output_name,dir=wd)}.{self.filetype}"
#                 )
#                 arguments.append(f"-O{output_path}")

#             with contextlib.nullcontext(structure) if isinstance(
#                 structure, str
#             ) else structure.tempfile(suffix=".xyz") as fp:
#                 input_filepath = mtr.expand(
#                     path=fp.name if hasattr(fp, "name") else fp, dir=wd
#                 )
#                 self.engine.execute(
#                     input_filepath=input_filepath,
#                     log_filepath=mtr.expand(path=self.log_name, dir=wd),
#                     arguments=arguments,
#                 )

#             if self.output_name is not None:
#                 with open(output_path, "r") as f:
#                     return "".join(f.readlines())


# class OpenbabelConvertToMol(OpenbabelConvertToFile):
#     def __init__(
#         self,
#         engine: mtr.OpenbabelEngine,
#         output_name: str,
#         log_name: Optional[str] = "obabel.log",
#         work_dir: Optional[str] = ".",
#         keep_logs: bool = True,
#         handlers: Optional[Iterable[mtr.Handler]] = None,
#         name: Optional[str] = None,
#     ) -> None:
#         super().__init__(
#             engine=engine,
#             filetype="mol",
#             output_name=output_name,
#             log_name=log_name,
#             work_dir=work_dir,
#             keep_logs=keep_logs,
#             handlers=handlers,
#             name=name,
#         )


# class OpenbabelConvertToPDB(OpenbabelConvertToFile):
#     def __init__(
#         self,
#         engine: mtr.OpenbabelEngine,
#         output_name: str,
#         log_name: Optional[str] = "obabel.log",
#         work_dir: Optional[str] = ".",
#         keep_logs: bool = True,
#         handlers: Optional[Iterable[mtr.Handler]] = None,
#         name: Optional[str] = None,
#     ) -> None:
#         super().__init__(
#             engine=engine,
#             filetype="pdb",
#             output_name=output_name,
#             log_name=log_name,
#             work_dir=work_dir,
#             keep_logs=keep_logs,
#             handlers=handlers,
#             name=name,
#         )


# class OpenbabelConvertToSDF(OpenbabelConvertToFile):
#     def __init__(
#         self,
#         engine: mtr.OpenbabelEngine,
#         output_name: str,
#         log_name: Optional[str] = "obabel.log",
#         work_dir: Optional[str] = ".",
#         keep_logs: bool = True,
#         handlers: Optional[Iterable[mtr.Handler]] = None,
#         name: Optional[str] = None,
#     ) -> None:
#         super().__init__(
#             engine=engine,
#             filetype="sdf",
#             output_name=output_name,
#             log_name=log_name,
#             work_dir=work_dir,
#             keep_logs=keep_logs,
#             handlers=handlers,
#             name=name,
#         )


# class OpenbabelConvertToSMILES(OpenbabelConvertToFile):
#     def __init__(
#         self,
#         engine: mtr.OpenbabelEngine,
#         output_name: str,
#         log_name: Optional[str] = "obabel.log",
#         work_dir: Optional[str] = ".",
#         keep_logs: bool = True,
#         handlers: Optional[Iterable[mtr.Handler]] = None,
#         name: Optional[str] = None,
#     ) -> None:
#         super().__init__(
#             engine=engine,
#             filetype="smi",
#             output_name=output_name,
#             log_name=log_name,
#             work_dir=work_dir,
#             keep_logs=keep_logs,
#             handlers=handlers,
#             name=name,
#         )

#     def run(self, structure: Union[str, mtr.Structure]) -> str:
#         super().run(structure=structure)

#         with open(
#             f"{mtr.expand(self.output_name,dir=self.work_dir)}.smi", "r"
#         ) as f:
#             smiles, *_ = re.search(r"([^\s]*).*", "".join(f.readlines())).groups()

#         return smiles



from __future__ import annotations
from typing import Any

import materia as mtr
import pickle
import rdkit, rdkit.Chem

__all__ = ["Molecule"]


def first(flist, default=None):
    """ Try each function in `flist` until one does not throw an exception, and
    return the return value of that function. If all functions throw exceptions,
    return `default` 

    Args: 
        flist - list of functions to try
        default - value to return if all functions fail

    Returns:
        return value of first function that does not throw exception, or
        `default` if all throw exceptions.

    TODO: Also accept a list of (f, (exceptions)) tuples, where f is the
    function as above and (exceptions) is a tuple of exceptions that f should
    expect. This allows you to still re-raise unexpected exceptions.
    """
    # from https://stackoverflow.com/a/13874877

    for f in flist:
        try:
            return f()
        except:
            continue
    else:
        return default


class Molecule:
    def __init__(self, structure) -> None:
        super().__setattr__("properties", {})
        if isinstance(structure, mtr.Structure):
            self.structure = structure
        elif isinstance(structure, str):
            self.structure = first(
                [
                    lambda: mtr.Structure.read(structure),
                    lambda: mtr.Structure.generate(smiles=structure),
                    lambda: mtr.Structure.retrieve(name=structure),
                ]
            )

        self.charge = rdkit.Chem.GetFormalCharge(self.structure.to_rdkit())
        self.multiplicity = (sum(self.structure.atomic_numbers) + self.charge) % 2 + 1

    def _from_file(self, structure):
        try:
            return mtr.Structure.read
        except:
            return False

    def save(self, filepath: str) -> None:
        """
        Pickle molecule to a given save file.

        Args:
            filepath: Path to file in which the molecule will be pickled. Can be an absolute or a relative path.
        """
        with open(mtr.expand(filepath), "wb") as f:
            pickle.dump(obj=self, file=f)

    @staticmethod
    def load(filepath: str) -> Molecule:
        """
        Load molecule from a pickle file.

        Args:
            filepath: Path to pickle file from which the molecule will be loaded. Can be an absolute or a relative path.

        Returns:
            Molecule retrieved from pickle file.

        """
        with open(mtr.expand(filepath), "rb") as f:
            mol = pickle.load(file=f)

        return mol

    def __getattr__(self, name: str) -> Any:
        if name == "properties":
            return self.properties
        try:
            return super().__getattribute__("properties")[name]
        except KeyError:
            return getattr(self.structure, name)

    def __setattr__(self, name: str, value) -> None:
        self.properties[name] = value



import itertools
import numpy as np
import scipy.linalg


class PointGroup:
    def __init__(self, products):
        self.products = products

    @property
    def labels(self):
        return tuple(set(self.products.values()))

    @property
    def order(self):
        return len(self.labels)

    @property
    def identity(self):
        identity_label = set.intersection(
            *(
                set((l1, l2))
                for (l1, l2), v in self.products.items()
                if l1 == v or l2 == v
            )
        ).pop()
        return GroupElement(label=identity_label, group=self)

    def inverse(self, label):
        inverse_label = next(
            l2 for (l1, l2), v in self.products.items() if l1 == label and v == 0
        )
        return GroupElement(label=inverse_label, group=self)

    def irreps(self):
        return set(
            tuple(np.real_if_close(X.trace()).item() for X in r)
            for r in self.regular_representation().reduce()
        )

    def cayley_table(self):
        lookup = dict(zip(self.labels, np.argsort(self.labels).argsort()))
        ctable = np.empty(
            shape=(self.order, self.order), dtype=int
        )  # FIXME: dtype should be 'S1' for single-character labels, etc.
        for (l1, l2), l3 in self.products.items():
            ctable[lookup[l1], lookup[l2]] = l3

        return ctable

    def regular_representation(self):
        return Representation(map=lambda g: g.regular_representation, group=self)

    def binary_operation(self, label1, label2):
        return self.products[(label1, label2)]

    def get_element(self, label):
        return GroupElement(label=label, group=self)

    @property
    def conjugacy_classes(self):
        return tuple(set(g.conjugacy_class for g in self))

    @property
    def conjugacy_representatives(self):
        return tuple(cls[0] for cls in self.conjugacy_classes)

    def __iter__(self):
        for l in self.labels:
            yield GroupElement(label=l, group=self)

    def __eq__(self, other):
        return hasattr(other, "products") and self.products == other.products

    def __hash__(self):
        return hash(frozenset(self.products.items()))


class GroupElement:
    def __init__(self, label, group):
        self.label = label
        self.group = group

    @property
    def regular_representation(self):
        permutation_dict = {
            l2: l3 for (l1, l2), l3 in self.group.products.items() if l1 == self.label
        }
        return self._permutation_matrix(permutation_dict=permutation_dict)

    @property
    def conjugacy_class(self):
        return tuple(set(g.inverse * self * g for g in self.group))

    def is_conjugate_to(self, other):
        return other in self.conjugacy_class

    @property
    def inverse(self):
        return self.group.inverse(label=self.label)

    def _permutation_matrix(self, permutation_dict):
        n = len(permutation_dict)
        I = np.eye(n, dtype=int)
        return np.hstack(
            [
                I[:, col_ind].reshape(n, -1)
                for _, col_ind in sorted(permutation_dict.items())
            ]
        )

    def __mul__(self, other):
        if (
            hasattr(other, "label")
            and hasattr(other, "group")
            and other.group == self.group
        ):
            return GroupElement(
                label=self.group.binary_operation(
                    label1=self.label, label2=other.label
                ),
                group=self.group,
            )

    def __eq__(self, other):
        return (
            hasattr(other, "label")
            and self.label == other.label
            and hasattr(other, "group")
            and self.group == other.group
        )

    def __hash__(self):
        return hash((self.label, self.group))


class Representation:
    def __init__(self, map, group):
        self.__call__ = map
        self.group = group

    def __iter__(self):
        for g in self.group:
            yield self.__call__(g)

    @property
    def dimension(self):
        try:
            (shape,) = tuple(set(rho.shape for rho in self))
        except ValueError:  # thrown from tuple unpacking into a single variable
            raise ValueError("Representation matrices have different shapes.")

        try:
            (dimension,) = tuple(set(shape))
        except ValueError:  # thrown from tuple unpacking into a single variable
            raise ValueError("Representation matrices are not square.")

        return dimension

    def reduce(self):
        # algorithm based on: http://sheaves.github.io/Group-Ring-Regular-Representation/
        H = self._commuting_matrix()
        _, P = np.linalg.eigh(H)
        blockified_reps = tuple(
            self._transform_rep(rho=rho, P=P) for rho in self
        )  # (self.representations[g] for g in list(self.representations.keys())[0].group.conjugacy_representatives))
        for reduced_reps in self.extract_blocks(*blockified_reps):
            map = lambda g: dict(zip((gp.label for gp in self.group), reduced_reps))[
                g.label
            ]
            rep = Representation(map=map, group=self.group)
            if rep.is_irreducible:
                yield rep
            else:
                yield from rep.reduce()

    @property
    def is_irreducible(self):
        return np.allclose(self._commuting_matrix(), np.eye(self.dimension))

    def _commuting_matrix(self):
        H_function_generator = (
            self.H_function(H=H) for H in self._hermitian_matrix_basis()
        )

        try:
            return next(
                (H for H in H_function_generator if not self._is_scalar_matrix(H=H))
            )
        except StopIteration:  # all H matrices are scalar matrices (so this is an irrep)
            return np.eye(self.dimension)

    def H_function(self, H):
        return sum(rho.T.conj() @ H @ rho for rho in self) / self.group.order

    def _hermitian_matrix_basis(self):
        for r, s in itertools.product(range(self.dimension), range(self.dimension)):
            E = np.zeros(shape=(self.dimension, self.dimension)).astype(np.complex)
            E[r, s] = 1j if r < s else 1
            yield E + E.T.conj() - np.diag(E.diagonal())

    def _is_scalar_matrix(self, H):
        return np.allclose(H, H[0, 0] * np.eye(self.dimension))

    def _transform_rep(self, rho, P):
        X = P.T.conj() @ rho @ P
        r = np.copy(X.real)
        i = np.copy(X.imag)
        r[np.isclose(r, 0)] = 0
        i[np.isclose(i, 0)] = 0
        return r + 1j * i

    def extract_blocks(self, *arrays):
        prev = -1
        (max_block_size,) = tuple(set(arrays[0].shape))
        for i in range(max_block_size - 1):
            if all(array[i + 1][i] == 0 and array[i][i + 1] == 0 for array in arrays):
                yield [array[prev + 1 : i + 1, prev + 1 : i + 1] for array in arrays]
                prev = i
        yield [array[prev + 1 : len(array), prev + 1 : len(array)] for array in arrays]

    # def _blockify(self, *matrices,first_time=True):
    #
    #     try:
    #         #FIXME: use np.ndim!!!
    #         (shape,) = tuple(set(A.shape for A in matrices))
    #     except ValueError:
    #         raise ValueError('Matrices must have same shapes.')
    #
    #     try:
    #         (dimension,) = tuple(set(shape))
    #     except ValueError:
    #         raise ValueError('Matrices must be square.')
    #
    #     Ts = (self.T_matrix(A=A) for A in matrices)
    #     null_vecs = scipy.linalg.null_space(sum(T.T@T for T in Ts)).T
    #     num_null_vecs,_ = null_vecs.shape
    #
    #     coeffs = np.random.uniform(low=1,high=2,size=num_null_vecs)
    #     coeffs /= np.linalg.norm(coeffs)
    #
    #     u = np.dot(coeffs,null_vecs)
    #     U = u.reshape(dimension,dimension)
    #
    #     X = U + U.T
    #     _,P = np.linalg.eigh(X)
    #
    #     blocked_matrices = [P.T@A@P for A in matrices]
    #     for B in blocked_matrices:
    #         B[np.isclose(B,0)] = 0
    #     if first_time:
    #         for list_of_mats in self.extract_blocks(*blocked_matrices):
    #             print(self._blockify(*list_of_mats,first_time=False))
    #     return blocked_matrices
    #
    # def T_matrix(self, A):
    #     try:
    #         (dimension,) = tuple(set(A.shape))
    #     except ValueError:
    #         raise ValueError('Cannot construct T-matrix for non-square matrix.')
    #
    #     if dimension == 1:
    #         return np.zeros(shape=(1,1))
    #     elif dimension == 2:
    #         a,b,c,d = np.ravel(A)
    #         return np.array([[0,-c,b,0],[-b,a-d,0,b],[c,0,d-a,-c],[0,a,-b,0]])
    #     elif dimension == 3:
    #         a,b,c,d,e,f,g,h,i = np.ravel(A)
    #         return np.array([[0,-d,-g,b,0,0,c,0,0],[-b,a-e,-h,0,b,0,0,c,0],
    #                          [-c,-f,a-i,0,0,b,0,0,c],[d,0,0,e-a,-d,-g,f,0,0],
    #                          [0,d,0,-b,0,-h,0,f,0],[0,0,d,-c,-f,e-i,0,0,f],
    #                          [g,0,0,h,0,0,i-a,-d,-g],[0,g,0,0,h,0,-b,i-e,-h],
    #                          [0,0,g,0,0,h,-c,-f,0]])
    #     else:
    #         raise ValueError('Cannot construct T-matrix for matrix with dimension greater than 3.')


class C1(PointGroup):
    def __init__(self):
        self.products = {(0, 0): 0}


class Cs(PointGroup):
    def __init__(self):
        self.products = {(0, 0): 0, (0, 1): 1, (1, 0): 1, (1, 1): 0}


class Ci(PointGroup):
    def __init__(self):
        self.products = {(0, 0): 0, (0, 1): 1, (1, 0): 1, (1, 1): 0}


class C2(PointGroup):
    def __init__(self):
        self.products = {(0, 0): 0, (0, 1): 1, (1, 0): 1, (1, 1): 0}


class C2(PointGroup):
    def __init__(self):
        self.products = {
            (0, 0): 0,
            (0, 1): 1,
            (1, 0): 1,
            (0, 2): 2,
            (2, 0): 2,
            (1, 1): 2,
            (1, 2): 0,
            (2, 1): 0,
            (2, 2): 1,
        }



from __future__ import annotations
from typing import Any, Dict, Iterable, Optional, Tuple, Union

import copy
import materia as mtr
import numpy as np
import scipy.integrate, scipy.interpolate

# import matplotlib.pyplot as plt
import warnings


class DataSeries:
    def __init__(self, x: mtr.Qty, y: mtr.Qty) -> None:
        self.x = x
        self.y = y

    # def plot(self):
    #     plt.plot(self.x.value, self.y.value)
    #     plt.show()


class DeltaSeries:
    def __init__(self, x: mtr.Qty, y: mtr.Qty) -> None:
        self.x = x
        self.y = y / x.unit

    def broaden(self, peak_func, x_eval, in_place=True):
        value = peak_func(x=self.x, y=self.y, x_eval=x_eval)

        if in_place:
            self.x = x_eval * self.x.unit
            self.y = value * self.y.unit
            return self
        else:
            new_series = copy.deepcopy(self)
            new_series.x = x_eval * self.x.unit
            new_series.y = value * self.y.unit
            return new_series

    # def plot(self):
    #     plt.scatter(self.x.value, self.y.value)
    #     plt.show()


class TimeSeries(DataSeries):
    @property
    def dt(self) -> mtr.Qty:
        # FIXME: should we even check for uniform spacing since this method,
        # like most/all methods, is susceptible to floating point errors?
        spacings = np.diff(self.x.value)
        dt, *_ = spacings
        if not np.allclose(spacings, dt):
            raise ValueError(
                "Time array does not have a unique spacing. Consider interpolating to a uniformly spaced time array first."
            )
        else:
            return dt * self.x.unit

    @property
    def T(self) -> mtr.Qty:
        return self.x[-1] - self.x[0]  # * self.x.unit

    def damp(self) -> None:  # , final_damp_value):
        final_damp_value = 1e-4  # Qty(value=1e-4,unit=self.x.unit)
        etasq = -np.log(final_damp_value) / self.T ** 2

        damp = np.exp(-(etasq * self.x ** 2).value)

        self.y *= damp

    def fourier_transform(self, pad_len=None):
        if pad_len is None:
            pad_len = len(self.x)

        fft_value = np.fft.fft(a=self.y.value, n=pad_len) / pad_len
        fft_real = fft_value.real * self.y.unit
        fft_imag = fft_value.imag * self.y.unit
        fft_freq = np.fft.fftfreq(n=pad_len, d=self.dt().value) / self.x.unit

        return (
            mtr.Spectrum(x=fft_freq, y=fft_real),
            mtr.Spectrum(x=fft_freq, y=fft_imag),
        )
        # self.x,convert_from=self.x_store_unit,convert_to='second')
        # frequencies = scipy.fftpack.fftfreq(n=pad_len,d=self.dt())
        # self.x = self.converter.convert(quantity=self.x,convert_from='second',convert_to=self.x_store_unit)

        # return ComplexSpectrum(x_vals=2*np.pi*frequencies,x_unit='radian_per_second',y=fft,y_unit=self.y_store_unit)


class Spectrum(DataSeries):
    def __init__(self, x, y):
        super().__init__(x=x, y=y)

    def match(self, match_to, in_place=True, interp_method="cubic_spline"):
        return self.extrapolate(x_extrap_to=match_to.x, in_place=in_place).interpolate(
            x_interp_to=match_to.x, in_place=in_place, method=interp_method
        )

    # simple linear extrapolation
    def extrapolate(self, x_extrap_to, in_place=True):
        if self.x.unit != x_extrap_to.unit:
            raise ValueError(
                "X-axis units do not match. Change units before extrapolating."
            )

        x_extrap, y_extrap = mtr.extrapolate(
            x=self.x.value, y=self.y.value, x_extrap_to=x_extrap_to.value
        )

        x_extrap = x_extrap * self.x.unit
        y_extrap = y_extrap * self.y.unit

        if in_place:
            self.x, self.y = x_extrap, y_extrap
            return self
        else:
            new_spectrum = copy.deepcopy(self)
            new_spectrum.x = x_extrap
            new_spectrum.y = y_extrap
            return new_spectrum

    def interpolate(self, x_interp_to, in_place=True, method="cubic_spline"):
        if self.x.unit != x_interp_to.unit:
            raise ValueError(
                "X-axis units do not match. Change units before interpolating."
            )

        x_interp, y_interp = mtr.interpolate(
            x=self.x.value, y=self.y.value, x_interp_to=x_interp_to.value, method=method
        )

        x_interp = x_interp * self.x.unit
        y_interp = y_interp * self.y.unit

        if in_place:
            self.x, self.y = x_interp, y_interp
            return self
        else:
            return mtr.Spectrum(x=x_interp, y=y_interp)

    # def plot(self, x=None):
    #     # FIXME: add axes labels, title
    #     if x is not None:
    #         plt.plot(x, self.y.match(match_to=x, in_place=False))
    #     else:
    #         plt.plot(self.x, self.y)
    #     plt.xlabel(f"Units: {self.x.unit}")
    #     plt.ylabel(f"Units: {self.y.unit}")
    #     plt.show()


# class ContinuousSpectrum(DataSeries):
#     def __init__(self, x, y):


class ComplexSpectrum(Spectrum):
    def __init__(self, x, y):
        super().__init__(x=x, y=y)

    # # FIXME: this code is stupid, figure out better syntax for real vs imag
    # def plot(self, real, imag, positive_frequencies_only=False):
    #     if positive_frequencies_only:
    #         plot_x = self.x[self.x >= 0]

    #     if real:
    #         plt.plot(plot_x, self.y.real)
    #     else:
    #         plt.plot(plot_x, self.y.imag)

    #     plt.show()


class PolarizabilitySpectrum(Spectrum):
    def __init__(self, x, y):
        super().__init__(x=x, y=y)


class AbsorptionSpectrum(Spectrum):
    def __init__(self, x, y):
        super().__init__(x=x, y=y)


class ReflectanceSpectrum(Spectrum):
    def __init__(self, x, y):
        super().__init__(x=x, y=y)

    def reflect_illuminant(self, illuminant):
        new_spectrum = copy.deepcopy(illuminant)
        new_spectrum.y = self.y * illuminant.match(match_to=self, in_place=False).y
        return new_spectrum


class TransmittanceSpectrum(Spectrum):
    def __init__(self, x, y):
        super().__init__(x=x, y=y)

    def transmit_illuminant(self, illuminant):
        new_spectrum = copy.deepcopy(illuminant)
        new_spectrum.y = self.y * illuminant.match(match_to=self, in_place=False).y
        return new_spectrum


class SPDSpectrum(Spectrum):
    def __init__(self, x, y):
        super().__init__(x=x, y=y)

        self.XYZ = self.tristimulus()
        self.X, self.Y, self.Z = self.XYZ

    def tristimulus(self):
        # FIXME: this is an ugly workaround to avoid circular import - change it!!
        from .cie_cmfs import (
            CIE1931ColorMatchingFunctionX,
            CIE1931ColorMatchingFunctionY,
            CIE1931ColorMatchingFunctionZ,
        )

        xbar = CIE1931ColorMatchingFunctionX().match(match_to=self, in_place=True)
        ybar = CIE1931ColorMatchingFunctionY().match(match_to=self, in_place=True)
        zbar = CIE1931ColorMatchingFunctionZ().match(match_to=self, in_place=True)

        X = scipy.integrate.simps(y=(self.y * xbar.y).value, x=self.x.value)
        Y = scipy.integrate.simps(y=(self.y * ybar.y).value, x=self.x.value)
        Z = scipy.integrate.simps(y=(self.y * zbar.y).value, x=self.x.value)

        return X, Y, Z

    # FIXME: verify correctness with test case
    def von_kries_XYZ(self, source_illuminant, destination_illuminant):
        source_lms = np.array(source_illuminant.LMS())
        destination_lms = np.array(destination_illuminant.LMS())
        LMS = np.array(self.LMS())[:, None]

        hpe = self.hunt_pointer_estevez_transform()
        adapted_LMS = np.diag(destination_lms / source_lms) @ LMS
        adapted_XYZ = np.linalg.inv(hpe) @ adapted_LMS

        adapted_X, adapted_Y, adapted_Z = adapted_XYZ.squeeze()

        return adapted_X, adapted_Y, adapted_Z

    def UVW(self, white_point=None):
        return self._XYZ_to_UVW(X=self.X, Y=self.Y, Z=self.Z, white_point=white_point)

    def von_kries_UVW(
        self, source_illuminant, destination_illuminant, white_point=None
    ):
        X, Y, Z = self.von_kries_XYZ(
            source_illuminant=source_illuminant,
            destination_illuminant=destination_illuminant,
        )

        return self._XYZ_to_UVW(
            X=X, Y=Y, Z=Z, white_point=white_point
        )  # FIXME: shouldn't the white_point actually be the desination illuminant's white point?

    def _XYZ_to_UVW(self, X, Y, Z, white_point=None):
        if white_point is None:
            U = 2 * X / 3
            V = Y
            W = 0.5 * (-X + 3 * Y + Z)
        else:
            u0, v0 = white_point
            x, y = self._XYZ_to_xy(X=X, Y=Y, Z=Z)

            u, v = self._xy_to_uv(x=x, y=y)

            W = 25 * np.power(Y, 1 / 3) - 17
            U = 13 * W * (u - u0)
            V = 13 * W * (v - v0)

        return U, V, W

    def LMS(self):
        hpe = self.hunt_pointer_estevez_transform()

        XYZ = np.array((self.XYZ))[:, None]
        L, M, S = (hpe @ XYZ).squeeze()

        return L, M, S

    def xy(self):
        return self._XYZ_to_xy(X=self.X, Y=self.Y, Z=self.Z)

    def _XYZ_to_xy(self, X, Y, Z):
        sum = X + Y + Z

        x = X / sum
        y = Y / sum

        return x, y

    def uv(self):
        x, y = self.xy()

        return self._xy_to_uv(x=x, y=y)

    def _xy_to_uv(self, x, y):
        denom = -2 * x + 12 * y + 3

        u = 4 * x / denom
        v = 6 * y / denom

        return u, v

    # color properties

    def CCT_DC(self):
        u, v = self.uv()
        uvT = self.planckian_locus_ucs(exact=True)

        def error(T):
            uT, vT = uvT(T=T)
            return (uT - u) ** 2 + (vT - v) ** 2

        CCT, DC_squared, _, _ = scipy.optimize.fminbound(
            func=error, x1=1000, x2=15000, full_output=True
        )

        return CCT, np.sqrt(DC_squared)

    def cri(self, strict=True):
        # FIXME: this is an ugly workaround to avoid circular import - change it!!
        from .cie_illuminants import CIEIlluminantDSeries
        from .cie_test_color_samples import (
            CIE1995TestColorSample01,
            CIE1995TestColorSample02,
            CIE1995TestColorSample03,
            CIE1995TestColorSample04,
            CIE1995TestColorSample05,
            CIE1995TestColorSample06,
            CIE1995TestColorSample07,
            CIE1995TestColorSample08,
        )

        CCT, DC = self.CCT_DC()

        if DC > 5.4e-3:
            if strict:
                raise ValueError(
                    "Distance from UCS Planckian locus too high. Illuminant is insufficiently white for accurate CRI determination."
                )
            else:
                warnings.warn(
                    "Distance from UCS Planckian locus too high. Illuminant is insufficiently white for accurate CRI determination."
                )

        if CCT < 5000:
            reference_illuminant = BlackbodySPD(T=CCT, normalize_to=1)
        else:
            reference_illuminant = CIEIlluminantDSeries(T=CCT, normalize_to=1)

        samples = (
            CIE1995TestColorSample01(),
            CIE1995TestColorSample02(),
            CIE1995TestColorSample03(),
            CIE1995TestColorSample04(),
            CIE1995TestColorSample05(),
            CIE1995TestColorSample06(),
            CIE1995TestColorSample07(),
            CIE1995TestColorSample08(),
        )

        u_ref, v_ref = reference_illuminant.uv()

        R_mean = np.mean(
            tuple(
                self.R_score(
                    test_illuminant=self,
                    reference_illuminant=reference_illuminant,
                    sample_reflectance=sample,
                )
                for sample in samples
            )
        )  # (TCS01,TCS02,TCS03,TCS04,TCS05,TCS06,TCS07,TCS08)))

        return R_mean

    def R_score(self, test_illuminant, reference_illuminant, sample_reflectance):
        u_ref, v_ref = reference_illuminant.uv()

        reflected_reference_illuminant = sample_reflectance.reflect_illuminant(
            illuminant=reference_illuminant
        )
        u_sample_ref, v_sample_ref = reflected_reference_illuminant.uv()
        X_sample_ref = reflected_reference_illuminant.X * 100 / reference_illuminant.Y
        Y_sample_ref = reflected_reference_illuminant.Y * 100 / reference_illuminant.Y
        Z_sample_ref = reflected_reference_illuminant.Z * 100 / reference_illuminant.Y
        U_ref, V_ref, W_ref = self._XYZ_to_UVW(
            X=X_sample_ref, Y=Y_sample_ref, Z=Z_sample_ref, white_point=(u_ref, v_ref)
        )

        reflected_test_illuminant = sample_reflectance.reflect_illuminant(
            illuminant=test_illuminant
        )
        u_sample_test, v_sample_test = reflected_test_illuminant.uv()
        u_sample_adapted, v_sample_adapted = self.von_kries_uv(
            u=u_sample_test,
            v=v_sample_test,
            source_illuminant=test_illuminant,
            destination_illuminant=reference_illuminant,
        )
        X_sample_test = reflected_test_illuminant.X * 100 / test_illuminant.Y
        Y_sample_test = reflected_test_illuminant.Y * 100 / test_illuminant.Y
        Z_sample_test = reflected_test_illuminant.Z * 100 / test_illuminant.Y
        U_test, V_test, W_test = self._XYZ_to_UVW(
            X=X_sample_test,
            Y=Y_sample_test,
            Z=Z_sample_test,
            white_point=(u_ref, v_ref),
        )
        # FIXME: using the next three lines improves the answer for high temp blackbodies but they seem wrong...
        # W_test = 25*np.power(Y_sample_test,1/3) - 17
        # U_test = 13*W_test*(u_sample_adapted - u_ref)
        # V_test = 13*W_test*(v_sample_adapted - v_ref)

        distance = np.sqrt(
            (U_ref - U_test) ** 2 + (V_ref - V_test) ** 2 + (W_ref - W_test) ** 2
        )

        return 100 - 4.6 * distance

    def von_kries_uv(self, u, v, source_illuminant, destination_illuminant):
        u_s, v_s = source_illuminant.uv()
        c_s = (4 - u_s - 10 * v_s) / v_s
        d_s = (1.708 * v_s + 0.404 - 1.481 * u_s) / v_s
        Y_s = source_illuminant.Y

        u_d, v_d = destination_illuminant.uv()
        c_d = (4 - u_d - 10 * v_d) / v_d
        d_d = (1.708 * v_d + 0.404 - 1.481 * u_d) / v_d
        Y_d = destination_illuminant.Y

        c = (4 - u - 10 * v) / v
        d = (1.708 * v + 0.404 - 1.481 * u) / v

        denom = 16.518 + 1.481 * c_d * c / c_s - d_d * d / d_s
        u_adapted = (10.872 + 0.404 * c_d * c / c_s - 4 * d_d * d / d_s) / denom
        v_adapted = 5.520 / denom

        return u_adapted, v_adapted

    # other standard quantities

    def planckian_locus_xyz(self, exact=False):
        if exact:
            # FIXME: this is an ugly workaround to avoid circular import - change it!!
            from .blackbody_spd import BlackbodySPD

            return lambda T: BlackbodySPD(T=T).xy()
        else:

            def x(T):
                if T >= 1667 and T <= 4000:
                    return (
                        -0.2661239e9 / T ** 3
                        - 0.2343589e6 / T ** 2
                        + 0.8776956e3 / T
                        + 0.179910
                    )
                elif T > 4000 and T <= 25000:
                    return (
                        -3.0258469e9 / T ** 3
                        + 2.1070379e6 / T ** 2
                        + 0.2226347e3 / T
                        + 0.240390
                    )

            def y(T):
                xc = x(T)
                if T >= 1667 and T <= 2222:
                    return (
                        -1.1063814 * xc ** 3
                        - 1.34811020 * xc ** 2
                        + 2.18555832 * xc
                        - 0.20219683
                    )
                elif T > 2222 and T <= 4000:
                    return (
                        -0.9549476 * xc ** 3
                        - 1.37418593 * xc ** 2
                        + 2.09137015 * xc
                        - 0.16748867
                    )
                elif T > 4000 and T <= 25000:
                    return (
                        3.0817580 * xc ** 3
                        - 5.87338670 * xc ** 2
                        + 3.75112997 * xc
                        - 0.37001483
                    )

            return lambda T: (x(T=T), y(T=T))

    def planckian_locus_ucs(self, exact=False):
        if exact:
            # FIXME: this is an ugly workaround to avoid circular import - change it!!
            from .blackbody_spd import BlackbodySPD

            return lambda T: BlackbodySPD(T=T).uv()
        else:

            def u(T):
                return (0.860117757 + 1.54118254e-4 * T + 1.28641212e-7 * T ** 2) / (
                    1 + 8.42420235e-4 * T + 7.08145163e-7 * T ** 2
                )

            def v(T):
                return (0.317398726 + 4.22806245e-5 * T + 4.20481691e-8 * T ** 2) / (
                    1 - 2.89741816e-5 * T + 1.61456053e-7 * T ** 2
                )

            return lambda T: (u(T=T), v(T=T))

    def hunt_pointer_estevez_transform(self):
        return np.array(
            [
                [0.4002400, 0.7076000, -0.0808100],
                [-0.2263000, 1.1653200, 0.0457000],
                [0.0000000, 0.0000000, 0.9182200],
            ]
        )


class RelativeSPDSpectrum(SPDSpectrum):
    def __init__(self, x, y, normalizing_x=560, normalize_to=1):
        [val,] = y[x.value == normalizing_x]
        super().__init__(
            x=x, y=normalize_to * y.value / val * mtr.unitless,
        )  # norm_factor*y/y.unit)

        self.XYZ = self.tristimulus()
        self.X, self.Y, self.Z = self.XYZ


# standard illuminants, default normalized to 100 @ 560 nm


class PhotopicResponse(RelativeSPDSpectrum):
    def __init__(self):
        # data taken from https://web.archive.org/web/20170131100357/http://files.cie.co.at/204.xls
        x = np.linspace(380, 780, 81) * mtr.nm

        V = [
            0.000039,
            0.000064,
            0.000120,
            0.000217,
            0.000396,
            0.000640,
            0.001210,
            0.002180,
            0.004000,
            0.007300,
            0.011600,
            0.016840,
            0.023000,
            0.029800,
            0.038000,
            0.048000,
            0.060000,
            0.073900,
            0.090980,
            0.112600,
            0.139020,
            0.169300,
            0.208020,
            0.258600,
            0.323000,
            0.407300,
            0.503000,
            0.608200,
            0.710000,
            0.793200,
            0.862000,
            0.914850,
            0.954000,
            0.980300,
            0.994950,
            1.000000,
            0.995000,
            0.978600,
            0.952000,
            0.915400,
            0.870000,
            0.816300,
            0.757000,
            0.694900,
            0.631000,
            0.566800,
            0.503000,
            0.441200,
            0.381000,
            0.321000,
            0.265000,
            0.217000,
            0.175000,
            0.138200,
            0.107000,
            0.081600,
            0.061000,
            0.044580,
            0.032000,
            0.023200,
            0.017000,
            0.011920,
            0.008210,
            0.005723,
            0.004102,
            0.002929,
            0.002091,
            0.001484,
            0.001047,
            0.000740,
            0.000520,
            0.000361,
            0.000249,
            0.000172,
            0.000120,
            0.000085,
            0.000060,
            0.000042,
            0.000030,
            0.000021,
            0.000015,
        ]
        y = np.array(V) * materia.unitless

        super().__init__(x=x, y=y, normalize_to=100)
